(* 書式指定付き文字列変換関数 *)

open OUnit
open Unix

(* カスタム数値書式指定文字列の構成要素 *)
type placeholder =
  | ZeroPlaceholder         (* ゼロプレースホルダ *)
  | DigitPlaceholder        (* 桁プレースホルダ *)
  | ImplicitPlaceholder     (* 残りの数値を格納する隠れたプレースホルダ *)
  | DecimalPoint            (* 小数点 *)
  | CommaSpecifier          (* カンマ記号 *)
  | PercentagePlaceholder   (* パーセントプレースホルダ *)
  | Character of char       (* 通常の文字 *)

(* 桁区切り記号の挿入を行った後の入力文字 *)
type input_char =
  | Digit of char           (* 通常の文字 *)
  | Comma                   (* 桁区切り記号 *)

(* 文字列を文字のリストに展開する。*)
let explode s =
  let l : char list ref = ref [] in
  if String.length s = 0 then
    !l
  else
    begin
      String.iter (fun c -> l := c :: !l) s;
      List.rev !l
    end

(* 文字のリストを結合して文字列とする。*)
let implode l =
  let b = Buffer.create (List.length l) in
  List.iter (Buffer.add_char b) l;
  Buffer.contents b

(* UNICODEをUTF-8文字コードに変換する。
  結果は，逆順（リトルエンディアン）の文字配列とする。*)
let utf8_of_point i =
  if i <= 0x007F then
    [char_of_int i]
  else if i <= 0x07FF then
    [char_of_int ((i land 0x3F) lor 0x80);
     char_of_int (((i lsr  6) land 0x1F) lor 0xC0)]
  else if i <= 0x0000FFFF then
    [char_of_int ((i land 0x3F) lor 0x80);
     char_of_int (((i lsr  6) land 0x3F) lor 0x80);
     char_of_int (((i lsr 12) land 0x0F) lor 0xE0)]
  else
    [char_of_int ((i land 0x3F) lor 0x80);
     char_of_int (((i lsr  6) land 0x3F) lor 0x80);
     char_of_int (((i lsr 12) land 0x3F) lor 0x80);
     char_of_int (((i lsr 18) land 0x07) lor 0xF0)]

(* 16進数文字を数値に変換する。*)
let hexval = function
  | '0'..'9' as c -> int_of_char c - int_of_char '0'
  | 'a'..'f' as c -> int_of_char c - int_of_char 'a' + 10
  | 'A'..'F' as c -> int_of_char c - int_of_char 'A' + 10
  | _ -> 0

(* エスケープ文字を解析する。
  結果は文字のリストで返す。*)
let parse_escape = function
  | [] -> [], []
  | 'b' :: cs -> ['\b'], cs
  | 'n' :: cs -> ['\n'], cs
  | 'r' :: cs -> ['\r'], cs
  | 't' :: cs -> ['\t'], cs
  | 'u' :: x0 :: x1 :: x2 :: x3 :: cs ->
      let i = 0x1000 * hexval x0 + 0x100 * hexval x1 + 0x10 * hexval x2 + hexval x3 in
      (utf8_of_point i), cs
  | c :: cs -> [c], cs

(* リテラル文字列を解析する。
  結果は，文字のリストで返す。*)
let rec parse_literal_string sep result = function
  | [] -> result, []
  | c :: cs when c = sep -> result, cs
  | '\\' :: cs ->
      let p, cs = parse_escape cs in
      parse_literal_string sep (p @ result) cs
  | c :: cs -> parse_literal_string sep (c :: result) cs

let to_char l = List.map (fun c -> Character c) l

(* 整数部を解析する。*)
let rec parse_integer_part result = function
  | [] -> (List.rev result, [])
  | '.' :: cs -> (List.rev result, '.' :: cs)
  | ';' :: cs -> (List.rev result, ';' :: cs)
  | '%' :: cs -> (List.rev result, '%' :: cs)
  | '0' :: cs -> parse_integer_part (ZeroPlaceholder :: result) cs
  | '#' :: cs -> parse_integer_part (DigitPlaceholder :: result) cs
  | ',' :: cs -> parse_integer_part (CommaSpecifier :: result) cs
  | '\\' :: cs ->
      let p, cs = parse_escape cs in
      parse_integer_part ((to_char p) @ result) cs
  | c :: cs when c = '\'' || c = '"' ->
      let p, cs = parse_literal_string c [] cs in
      parse_integer_part ((to_char p) @ result) cs
  | c :: cs -> parse_integer_part (Character c :: result) cs

(* 小数部を解析する。*)
let rec parse_fraction_part result = function
  | [] -> (List.rev result, [])
  | ';' :: cs -> (List.rev result, cs)
  | '.' :: cs -> parse_fraction_part (DecimalPoint :: result) cs
  | '0' :: cs -> parse_fraction_part (ZeroPlaceholder :: result) cs
  | '#' :: cs -> parse_fraction_part (DigitPlaceholder :: result) cs
  | ',' :: cs -> parse_fraction_part (CommaSpecifier :: result) cs
  | '%' :: cs -> parse_integer_part (PercentagePlaceholder :: result) cs
  | '\\' :: cs ->
      let p, cs = parse_escape cs in
      parse_integer_part ((to_char p) @ result) cs
  | c :: cs -> parse_fraction_part (Character c :: result) cs

(* 最初のゼロ/桁プレースホルダの直前に暗黙的プレースホルダを挿入する。
  文字列を整形する際に，余った数字を暗黙的プレースホルダの位置に挿入する。*)
let rec insert_implicit_placeholder result = function
  | [] -> List.rev result
  | p :: ps when p = DigitPlaceholder || p = ZeroPlaceholder ->
      (List.rev result) @ (ImplicitPlaceholder :: p :: ps)
  | p :: ps -> insert_implicit_placeholder (p :: result) ps

(* カンマ記号の数をカウントする。*)
let rec count_comma a s = function
  | [] -> a, s
  | CommaSpecifier :: ps -> count_comma (a + 1) (s + 1) ps
  | ImplicitPlaceholder :: ps -> count_comma 0 0 ps
  | ZeroPlaceholder :: ps -> count_comma a 0 ps
  | DigitPlaceholder :: ps -> count_comma a 0 ps
  | p :: ps -> count_comma a s ps

(* パーセントプレースホルダを含んでいるか？*)
let rec has_percentage = function
  | [] -> false
  | PercentagePlaceholder :: ps -> true
  | p :: ps -> has_percentage ps

(* 小数点部の精度（桁数）をカウントする。*)
let count_placeholder ps =
  let f n p =
    if p = ZeroPlaceholder || p = DigitPlaceholder then
      n + 1
    else
      n
  in
  List.fold_left f 0 ps

(* セクションを解析する。*)
let parse_section result cs =
  let i, cs = parse_integer_part [] cs in
  let f, cs = parse_fraction_part [] cs in
  let i = insert_implicit_placeholder [] i in
  let num_comma, num_scaling = count_comma 0 0 i in
  let has_thousand = num_comma - num_scaling > 0 in
  let num_percentage = if has_percentage f then 1 else 0 in
  let scaling = 3 * num_scaling - 2 * num_percentage in
  (i, f, has_thousand, scaling), cs

(* カスタム数値書式指定文字列をパースする。*)
let parse_format s =
  let rec parse_inner result = function
    | [] -> List.rev result
    | cs ->
        let r, cs = parse_section [] cs in
        parse_inner (r :: result) cs
  in
  let fmt = parse_inner [] (explode s) in
  fmt

(* 文字リストを'.'で二つに分割する。*)
let separate_char_list l =
  let rec separate_char_list_inner head = function
    | [] -> (head, [])
    | '.' :: rest -> (head, rest)
    | c :: rest -> separate_char_list_inner (c :: head) rest
  in
  let h, t = separate_char_list_inner [] l in
  (List.rev h, t)

(* 入力数値をパースする。
 ・入力が１桁"0"であれば，空文字列と解釈する。
 ・整数部と小数点部に分割する。
 ・整数部は，逆順のリストとする。
 ・整数部には，３桁単位に区切り記号(Camma)を挿入する。
*)
let parse_input i thousand =
  let rec parse_head = function
    | [] -> []
    | c :: rest -> Digit c :: parse_head rest
  in
  let rec parse_tail = function
    | [] -> []
    | c :: rest -> Digit c :: parse_tail rest
  in
  let rec separate_thousand = function
    | [] -> []
    | Digit d1 :: Digit d2 :: Digit d3 :: rest when List.length rest > 0 ->
        Digit d1 :: Digit d2 :: Digit d3 :: Comma :: (separate_thousand rest)
    | d :: rest -> d :: separate_thousand rest
  in
  let l = explode (if i = "0" then "" else i) in
  let head, tail = separate_char_list l in
  let parsed_head = parse_head (List.rev head) in
  let parsed_head = if thousand then separate_thousand parsed_head else parsed_head in
  (parsed_head, parse_tail tail)

(* 整数部の整形を行う。*)
let rec format_head input = function
  | [] -> []
  | ZeroPlaceholder :: form_rest ->
      begin match input with
      | [] -> '0' :: format_head [] form_rest
      | Digit d :: in_rest -> d :: format_head in_rest form_rest
      | Comma :: in_rest -> ',' :: format_head in_rest (ZeroPlaceholder :: form_rest)
      end
  | DigitPlaceholder :: form_rest ->
      begin match input with
      | [] -> format_head [] form_rest
      | Digit d :: in_rest -> d :: format_head in_rest form_rest
      | Comma :: in_rest -> ',' :: format_head in_rest (ZeroPlaceholder :: form_rest)
      end
  | Character c :: form_rest -> c :: format_head input form_rest 
  | ImplicitPlaceholder :: form_rest ->
      begin match input with
      | [] -> format_head [] form_rest
      | Digit d :: in_rest -> d :: format_head in_rest (ImplicitPlaceholder :: form_rest)
      | Comma :: in_rest -> ',' :: format_head in_rest (ImplicitPlaceholder :: form_rest)
      end
  | _ :: form_rest -> format_head input form_rest

(* 小数部の整形を行う。*)
let format_tail input fmt =
  let rec format_tail_inner input = function
    | [] -> []
    | ZeroPlaceholder :: form_rest ->
        begin match input with
        | [] -> '0' :: format_tail_inner [] form_rest
        | Digit d :: in_rest -> d :: format_tail_inner in_rest form_rest
        | Comma :: in_rest -> format_tail_inner in_rest (ZeroPlaceholder :: form_rest)
        end
    | DigitPlaceholder :: form_rest ->
        begin match input with
        | [] -> format_tail_inner [] form_rest
        | Digit d :: in_rest -> d :: format_tail_inner in_rest form_rest
        | Comma :: in_rest -> format_tail_inner in_rest (ZeroPlaceholder :: form_rest)
        end
    | Character c :: form_rest -> c :: format_tail_inner input form_rest 
    | DecimalPoint :: form_rest -> '.' :: format_tail_inner input form_rest
    | PercentagePlaceholder :: form_rest -> '%' :: format_tail_inner input form_rest
    | _ :: form_rest -> format_tail_inner input form_rest
  in
  format_tail_inner input fmt

(* mで指定した桁位置で四捨五入する。*)
let round i m =
  let factor = 0.5 /. 10.0 ** (float_of_int m) in
  if i >= 0.0 then
    i +. factor
  else
    -. (abs_float i +. factor)

(* 浮動小数点を表示するのに必要な桁数を求める。*)
let count_fraction s =
  let f n c = if c = '.' then 0 else n + 1 in
  List.fold_left f 0 (explode s)

let select_fmt fmt n =
  if (List.length fmt) = 3 then
    begin
      if n > 0. then
        List.hd fmt
      else if n < 0. then
        List.nth fmt 1
      else
        List.nth fmt 2
    end
  else if (List.length fmt) = 2 then
    begin
      if n >= 0. then
        List.hd fmt
      else
        List.nth fmt 1
    end
  else
    begin
      List.hd fmt
    end

(* 整数を書式文字列に応じて文字列に変換する。 *)
let format_int n s =
  let fmt = parse_format s in
  let hform, tform, thousand, scale = select_fmt fmt (float_of_int n) in
  let n = if List.length fmt > 1 then abs n else n in
  let num = string_of_float (round ((float_of_int n) /. 10.0 ** (float_of_int scale)) 0) in
  let hinput, tinput = parse_input num thousand in
  let head = List.rev (format_head hinput (List.rev hform)) in
  let tail = format_tail [] tform in
  implode head ^ implode tail

(* 実数を書式文字列に応じて文字列に変換する。*)
let format_float e s =
  let fmt = parse_format s in
  let hform, tform, thousand, scale = select_fmt fmt e in
  let e = if List.length fmt > 1 then abs_float e else e in
  let scaled_num = e /. 10.0 ** (float_of_int scale) in
  let num_fraction = count_fraction (string_of_float scaled_num) in
  let num_placeholder = count_placeholder tform in
  let num = string_of_float
    (if num_fraction > num_placeholder then
      round (scaled_num) num_placeholder
    else
      scaled_num)
  in
  let hinput, tinput = parse_input num thousand in
  let head = List.rev (format_head hinput (List.rev hform)) in
  let tail = format_tail tinput tform in
  implode head ^ implode tail

(* Zellerの公式により曜日を求める。*)
let wday_of_date y m d =
  let y, m = if m <= 2 then y - 1, m + 12 else y, m in
  (y + (y / 4) - (y / 100) + (y / 400) + (13 * m + 8) / 5 + d) mod 7

(* ユリウス通日を求める。*)
let julian_day_of_date y m d =
  let y, m = if m <= 2 then y - 1, m + 12 else y, m in
  let d =
    if (y * 12 + m) * 31 + d >= (1582 * 12 + 10) * 31 + 15 then
      d + 2 - y / 100 + y / 400
    else
      d
  in
  (int_of_float (365.25 *. (float_of_int (y + 4716)))) +
  (int_of_float (30.6001 *. (float_of_int (m + 1)))) +
  (d - 1524)

(* 年初からの経過日を求める。*)
let yday_of_date y m d =
 (julian_day_of_date y m d) - (julian_day_of_date y 1 1)


let enum_of_B = ["Jan."; "Feb."; "Mar."; "Apr."; "May"; "Jun.";
                 "Jul."; "Aug."; "Sep."; "Oct."; "Nov."; "Dec."]

let enum_of_BB = ["１月"; "２月"; "３月"; "４月"; "５月"; "６月";
                 "７月"; "８月"; "９月"; "１０月"; "１１月"; "１２月"]

let enum_of_A = ["日"; "月"; "火"; "水"; "木"; "金"; "土"]
let enum_of_AA = ["日曜日"; "月曜日"; "火曜日"; "水曜日"; "木曜日"; "金曜日"; "土曜日"]

let enum_of_G = ["A.D."; "M"; "T"; "S"; "H"]
let enum_of_GG = ["西暦"; "明治"; "大正"; "昭和"; "平成"]

let enum_of_t = ["AM"; "PM"]
let enum_of_tt = ["午前"; "午後"]

let string_of_YY tm = Printf.sprintf "%02d" ((1900 + tm.tm_year) mod 100)
let string_of_YYYY tm = Printf.sprintf "%04d" (1900 + tm.tm_year)

let string_of_M tm = Printf.sprintf "%d" (tm.tm_mon + 1)
let string_of_MM tm = Printf.sprintf "%02d" (tm.tm_mon + 1)

let string_of_B ?(enum=enum_of_B) tm = List.nth enum tm.tm_mon
let string_of_BB ?(enum=enum_of_BB) tm = List.nth enum tm.tm_mon

let string_of_D tm = Printf.sprintf "%d" tm.tm_mday
let string_of_DD tm = Printf.sprintf "%02d" tm.tm_mday

let string_of_A ?(enum=enum_of_A) tm = List.nth enum tm.tm_wday
let string_of_AA ?(enum=enum_of_AA) tm = List.nth enum tm.tm_wday

let g_value tm =
  let year = 1900 + tm.tm_year in
  let date = 10000 * year + 100 * (tm.tm_mon + 1) + tm.tm_mday in
  if date < 18680908 then
    0, year
  else if date < 19120730 then 
    1, year - 1867
  else if date < 19261225 then 
    2, year - 1911
  else if date < 19890108 then 
    3, year - 1925
  else
    4, year - 1988

let string_of_G ?(enum=enum_of_G) tm =
  let g, _e = g_value tm in
  List.nth enum g

let string_of_GG ?(enum=enum_of_GG) tm =
  let g, _e = g_value tm in
  List.nth enum g

let string_of_E tm =
  let _g, e = g_value tm in
  Printf.sprintf "%d" e

let string_of_EE tm =
  let _g, e = g_value tm in
  Printf.sprintf "%02d" e

let h12_of_h24 h = if h > 12 then h - 12 else h
let string_of_h tm = Printf.sprintf "%d" tm.tm_hour
let string_of_hh tm = Printf.sprintf "%02d" tm.tm_hour
let string_of_H tm = Printf.sprintf "%d" (h12_of_h24 tm.tm_hour)
let string_of_HH tm = Printf.sprintf "%02d" (h12_of_h24 tm.tm_hour)

let string_of_m tm = Printf.sprintf "%d" tm.tm_min
let string_of_mm tm = Printf.sprintf "%02d" tm.tm_min

let string_of_s tm = Printf.sprintf "%d" tm.tm_sec
let string_of_ss tm = Printf.sprintf "%02d" tm.tm_sec

let t_value h = if h < 12 then 0 else 1
let string_of_t ?(e=enum_of_t) tm = List.nth e (t_value tm.tm_mday)
let string_of_tt ?(e=enum_of_tt) tm = List.nth e (t_value tm.tm_mday)

let string_of_char ch = Printf.sprintf "%c" ch

(* DateTime書式指定を解析する。*)
let format_tm tm s =
  let concat l =
    let len = List.fold_left (fun len s -> len + String.length s) 0 l in
    let b = Buffer.create len in
    List.iter (Buffer.add_string b) l;
    Buffer.contents b
  in
  let rec parse_format = function
    | [] -> []
    | 'Y'::'Y'::'Y'::'Y'::rest ->
        string_of_YYYY tm :: (parse_format rest)
    | 'Y'::'Y'::rest -> string_of_YY tm :: (parse_format rest)
    | 'M'::'M'::rest -> string_of_MM tm :: (parse_format rest)
    | 'M'::rest      -> string_of_M tm :: (parse_format rest)
    | 'B'::'B'::rest -> string_of_BB tm :: (parse_format rest)
    | 'B'::rest      -> string_of_B tm :: (parse_format rest)
    | 'D'::'D'::rest -> string_of_DD tm :: (parse_format rest)
    | 'D'::rest      -> string_of_D tm :: (parse_format rest)
    | 'A'::'A'::rest -> string_of_AA tm :: (parse_format rest)
    | 'A'::rest      -> string_of_A tm :: (parse_format rest)
    | 'G'::'G'::rest -> string_of_GG tm :: (parse_format rest)
    | 'G'::rest      -> string_of_G tm :: (parse_format rest)
    | 'E'::'E'::rest -> string_of_EE tm :: (parse_format rest)
    | 'E'::rest      -> string_of_E tm :: (parse_format rest)
    | 'h'::'h'::rest -> string_of_hh tm :: (parse_format rest)
    | 'h'::rest      -> string_of_h tm :: (parse_format rest)
    | 'H'::'H'::rest -> string_of_HH tm :: (parse_format rest)
    | 'H'::rest      -> string_of_H tm :: (parse_format rest)
    | 'm'::'m'::rest -> string_of_mm tm :: (parse_format rest)
    | 'm'::rest      -> string_of_m tm :: (parse_format rest)
    | 's'::'s'::rest -> string_of_ss tm :: (parse_format rest)
    | 's'::rest      -> string_of_s tm :: (parse_format rest)
    | 't'::'t'::rest -> string_of_tt tm :: (parse_format rest)
    | 't'::rest      -> string_of_t tm :: (parse_format rest)
    | '\\'::rest ->
        let chars, rest = parse_escape rest in
        implode (List.rev chars) :: (parse_format rest)
    | ch::rest when ch = '\'' || ch = '"' ->
        let chars, rest = parse_literal_string ch [] rest in
        implode (List.rev chars) :: (parse_format rest)
    | ch::rest ->
        string_of_char ch :: (parse_format rest)
  in
  concat (parse_format (explode s))


(* ISO 8601形式時刻文字列をtmに変換する。*)
let tm_of_string s =
  let is_digit c = '0' <= c && c <= '9' in
  let is_digits l = List.for_all is_digit l in
  let int_of_digits l = int_of_string (implode l) in
  let parse_ss = function
    | [] -> [0]
    | ':' :: s0 :: s1 :: _ when is_digits [s0; s1] ->
        (int_of_digits [s0; s1]) :: []
    | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))
  in
  let parse_mmss = function
    | [] -> [0; 0]
    | ':' :: m0 :: m1 :: ss when is_digits [m0; m1] ->
        (int_of_digits [m0; m1]) :: parse_ss ss
    | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))
  in
  let parse_hhmmss = function
    | h0 :: h1 :: mmss when is_digits [h0; h1] ->
        (int_of_digits [h0; h1]) :: parse_mmss mmss
    | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))
  in
  let parse_dd = function
    | [] -> [1; 0; 0; 0]
    | '-' :: d0 :: d1 :: sep :: hhmmss
      when is_digits [d0; d1] && (sep = 'T' || sep = ' ') ->
        (int_of_digits [d0; d1]) :: parse_hhmmss hhmmss
    | sep :: hhmmss when sep = 'T' || sep = ' ' ->
        1 :: parse_hhmmss hhmmss
    | '-' :: d0 :: d1 :: [] when is_digits [d0; d1] ->
        (int_of_digits [d0; d1]) :: [0; 0; 0]
    | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))
  in
  let parse_mmdd = function
    | [] -> [1; 1; 0; 0; 0]
    | '-' :: m0 :: m1 :: dd when is_digits [m0; m1] ->
        (int_of_digits [m0; m1]) :: parse_dd dd
    | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))
  in
  let tm =
    match explode s with
    | y0 :: y1 :: y2 :: y3 :: mmdd when is_digits [y0; y1; y2; y3] ->
        (int_of_digits [y0; y1; y2; y3]) :: parse_mmdd mmdd
    | hhmmss ->
        [0; 0; 0] @ (parse_hhmmss hhmmss)
  in
  match tm with
  | year :: mon :: day :: hour :: min :: sec :: [] ->
      {tm_sec = sec; tm_min = min; tm_hour = hour;
       tm_mday = day; tm_mon = mon - 1; tm_year = year - 1900;
       tm_wday = (wday_of_date year mon day);
       tm_yday = (yday_of_date year mon day);
       tm_isdst = false}
  | _ -> raise (Invalid_argument ("illegal datetime format: " ^ s))

(* tmをISO 8601形式時刻文字列に変換する。*)
let string_of_tm tm =
  format_tm tm "YYYY-MM-DDThh:mm:ss"

(* TEST *)

let test_custom =
  "Custom" >:::
    [
      "format_int #1" >:: (fun () ->
        "1.1" @? (format_int 12345 "00000000" = "00012345");
        "1.2" @? (format_int 12345 "000" = "12345");
        "1.3" @? (format_int 12345 "0.00000" = "12345.00000");
        "2.1" @? (format_int 12345 "########" = "12345");
        "2.2" @? (format_int 12345 "###" = "12345");
        "2.3" @? (format_int 12345 "#.#####" = "12345.");
        "2.4" @? (format_int 1234567890 "(###)###-####" = "(123)456-7890");
        "3.1" @? (format_int 1234567890 "[000]-[####]-[####]" = "[012]-[3456]-[7890]");
        "4.1" @? (format_int 1234567890 "#,#" = "1,234,567,890");
        "4.2" @? (format_int 1234567890 "#,,#" = "1,234,567,890");
        "4.3" @? (format_int 1234567890 "#," = "1234568");
        "4.4" @? (format_int 1234567890 "#,," = "1235");
        "4.5" @? (format_int (-1234567890) "#," = "-1234568");
        "4.6" @? (format_int 1234567890 "#,##," = "1,234,568");
        "4.7" @? (format_int 100000 "#,#" = "100,000");
      );

      "format_int #2" >:: (fun () ->
        "1.1" @? (format_int 123 "+###;-###" = "+123");
        "1.2" @? (format_int 123 "０以上;０未満" = "０以上");
        "1.3" @? (format_int (-123) "+###;-###" = "-123");
        "1.4" @? (format_int (-123) "０以上;０未満" = "０未満");
        "1.5" @? (format_int 0 "正;負;０" = "０");
      );

      "format_float #1" >:: (fun () ->
        "1.1" @? (format_float 123.45 "00000.0" = "00123.5");
        "2.1" @? (format_float 123.45 "#####.#" = "123.5");
        "2.2" @? (format_float 123.45 "#.#####" = "123.45");
        "3.1" @? (format_float 0.30944 "0.##%" = "30.94%");
      );

      "MSDN" >:: (fun () ->
        "1.1" @? (format_int 123 "#####" = "123");
        "1.2" @? (format_int 123 "00000" = "00123");
        "1.3" @? (format_int 1234567890 "(###)###-####" = "(123)456-7890");
        "2.1" @? (format_float 1.2 "#.##" = "1.2");
        "2.2" @? (format_float 1.2 "0.00" = "1.20");
        "2.3" @? (format_float 1.2 "00.00" = "01.20");
        "3.1" @? (format_int 1234567890 "#,#" = "1,234,567,890");
        "3.2" @? (format_int 1234567890 "#,," = "1235");
        "3.3" @? (format_int 1234567890 "#,,," = "1");
        "3.4" @? (format_int 1234567890 "#,##0,," = "1,235");
        "4.1" @? (format_float 0.086 "#0.##%" = "8.6%");
        "4.2" @? (format_int 123456 "[##-##-##]" = "[12-34-56]");
        "4.3" @? (format_int 1234 "##;(##)" = "1234");
        "4.4" @? (format_int (-1234) "##;(##)" = "(1234)");
      );

      "tm_of_string #1" >:: (fun () ->
        "1.1" @? (match tm_of_string "2010" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 0 && mday = 1 && hour = 0 && min = 0 && sec = 0);
        "1.2" @? (match tm_of_string "2010-09" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 1 && hour = 0 && min = 0 && sec = 0);
        "1.3" @? (match tm_of_string "2010-09-23" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 23 && hour = 0 && min = 0 && sec = 0);
        "1.4" @? (match tm_of_string "2010-09-23T09" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 23 && hour = 9 && min = 0 && sec = 0);
        "1.5" @? (match tm_of_string "2010-09-23T09:58" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 23 && hour = 9 && min = 58 && sec = 0);
        "1.6" @? (match tm_of_string "2010-09-23T09:58:12" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 23 && hour = 9 && min = 58 && sec = 12);
        "1.7" @? (match tm_of_string "2010-09-23 09:58:12" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 23 && hour = 9 && min = 58 && sec = 12);
        "1.8" @? (match tm_of_string "2010-09T09:58:12" with
          {tm_year = year; tm_mon = mon; tm_mday = mday;
           tm_hour = hour; tm_min = min; tm_sec = sec}
            -> year = 110 && mon = 8 && mday = 1 && hour = 9 && min = 58 && sec = 12);
      );
      
      "tm_of_string #2" >:: (fun () ->
        "1.1" @? (match tm_of_string "1965-06-03" with
          {tm_wday = wday} -> wday = 4);
        "1.2" @? (match tm_of_string "2010-01-01" with
          {tm_yday = yday} -> yday = 0);
        "1.3" @? (match tm_of_string "2010-12-31" with
          {tm_yday = yday} -> yday = 364);
        "1.4" @? (match tm_of_string "2008-12-31" with
          {tm_yday = yday} -> yday = 365);
      );
      
      "tm_of_string #3" >:: (fun () ->
        "1.1" @? (try ignore (tm_of_string "20100923"); false with _ -> true);
        "1.2" @? (try ignore (tm_of_string "2010.09.23"); false with _ -> true);
        "1.3" @? (try ignore (tm_of_string "2010-09-23t12:34"); false with _ -> true);
        "1.4" @? (try ignore (tm_of_string "2010-9-23"); false with _ -> true);
        "1.5" @? (try ignore (tm_of_string "2010-09-23T9:30"); false with _ -> true);
        "1.6" @? (try ignore (tm_of_string "xxx"); false with _ -> true);
      );

      "format_tm #1" >:: (fun () ->
        let tm = tm_of_string "2010-09-03T08:03:05" in
        "1.1" @? (format_tm tm "YY" = "10");
        "1.2" @? (format_tm tm "YYYY" = "2010");
        "2.1" @? (format_tm tm "M" = "9");
        "2.2" @? (format_tm tm "MM" = "09");
        "3.1" @? (format_tm tm "B" = "Sep.");
        "3.2" @? (format_tm tm "BB" = "９月");
        "4.1" @? (format_tm tm "D" = "3");
        "4.2" @? (format_tm tm "DD" = "03");
        "5.1" @? (format_tm tm "A" = "金");
        "5.2" @? (format_tm tm "AA" = "金曜日");
        "6.1" @? (format_tm tm "G" = "H");
        "6.2" @? (format_tm tm "GG" = "平成");
        "7.1" @? (format_tm tm "E" = "22");
        "7.2" @? (format_tm tm "EE" = "22");
        "8.1" @? (format_tm tm "h" = "8");
        "8.2" @? (format_tm tm "hh" = "08");
        "9.1" @? (format_tm tm "H" = "8");
        "9.2" @? (format_tm tm "HH" = "08");
        "10.1" @? (format_tm tm "m" = "3");
        "10.2" @? (format_tm tm "mm" = "03");
        "11.1" @? (format_tm tm "s" = "5");
        "11.2" @? (format_tm tm "ss" = "05");
        "12.1" @? (format_tm tm "t" = "AM");
        "12.2" @? (format_tm tm "tt" = "午前");
      );

      "format_tm #2" >:: (fun () ->
        let tm = tm_of_string "1965-11-23T15:29:45" in
        "1.1" @? (format_tm tm "YY" = "65");
        "1.2" @? (format_tm tm "YYYY" = "1965");
        "2.1" @? (format_tm tm "M" = "11");
        "2.2" @? (format_tm tm "MM" = "11");
        "3.1" @? (format_tm tm "B" = "Nov.");
        "3.2" @? (format_tm tm "BB" = "１１月");
        "4.1" @? (format_tm tm "D" = "23");
        "4.2" @? (format_tm tm "DD" = "23");
        "5.1" @? (format_tm tm "A" = "火");
        "5.2" @? (format_tm tm "AA" = "火曜日");
        "6.1" @? (format_tm tm "G" = "S");
        "6.2" @? (format_tm tm "GG" = "昭和");
        "7.1" @? (format_tm tm "E" = "40");
        "7.2" @? (format_tm tm "EE" = "40");
        "8.1" @? (format_tm tm "h" = "15");
        "8.2" @? (format_tm tm "hh" = "15");
        "9.1" @? (format_tm tm "H" = "3");
        "9.2" @? (format_tm tm "HH" = "03");
        "10.1" @? (format_tm tm "m" = "29");
        "10.2" @? (format_tm tm "mm" = "29");
        "11.1" @? (format_tm tm "s" = "45");
        "11.2" @? (format_tm tm "ss" = "45");
        "12.1" @? (format_tm tm "t" = "PM");
        "12.2" @? (format_tm tm "tt" = "午後");
      );

      "format_tm #3" >:: (fun () ->
        "1.1" @? (format_tm (tm_of_string "1902") "GGEE" = "明治35");
        "1.2" @? (format_tm (tm_of_string "1918") "GGEE" = "大正07");
        "1.3" @? (format_tm (tm_of_string "1926-12-24") "GE" = "T15");
        "1.4" @? (format_tm (tm_of_string "1926-12-25") "GGE" = "昭和1");
        "1.5" @? (format_tm (tm_of_string "1868-09-08") "GGEE" = "明治01");
        "1.5" @? (format_tm (tm_of_string "1868-01-01") "GGEE" = "西暦1868");
      );
    ]

let test () =
  OUnit.run_test_tt (OUnit.TestList [test_custom])


