(** 書式指定付き文字列変換ライブラリ

  書式指定文字列の指示にしたがって，数値/日付/時刻を文字列に変換します。

  @author Kenji Kajita, Field Works, LLC.
  @version 0.2
*)

(** {6 数値の書式指定}
  MS Officeもしくは.Net Framework カスタム数値書式指定文字列に類似した書式指
  定文字列を使用します。

  {10 【利用可能な書式指定文字】}
  - [0 ]         ゼロプレースホルダ
  - [# ]         桁プレースホルダ
  - [. ]         小数点
  - [, ]         桁区切り記号，値の位取り
  - [% ]         パーセントプレースホルダ
  - [\<文字> ]   エスケープ文字
  - ['<文字列>', "<文字列>" ]    リテラル文字列
  - [; ]         セクション区切り記号（最大3セクション：正;負;ゼロ）
  - [その他の文字 ]      結果の文字列にコピーされます。

  ※指数表記(E0, E+0, e-00,...)には未対応です。

  {10 【利用可能なエスケープ文字】}
  - [\b ]            バックスペース(BS)
  - [\n ]            行送り(LF)
  - [\r ]            復帰(CR)
  - [\t ]            水平タブ(TAB)
  - [\u<HHHH> ]      16進数値[HHHH]を持つUnicode文字
  - [\<文字> ]       文字自身

  {10 【使用例】}
  - [format_int 123 "#####" -> "123"]
  - [format_int 123 "00000" -> "00123"]
  - [format_int 1234567890 "(###)###-####" -> "(123)456-7890"]
  - [format_float 1.2 "#.##" -> "1.2"]
  - [format_float 1.2 "0.00" -> "1.20"]
  - [format_float 1.2 "00.00" -> "01.20"]
  - [format_int 1234567890 "#,#" -> "1,234,567,890"]
  - [format_int 1234567890 "#,," -> "1235"]
  - [format_int 1234567890 "#,,," -> "1"]
  - [format_int 1234567890 "#,##0,," -> "1,235"]
  - [format_float 0.086 "#0.##%" -> "8.6%"]
  - [format_int 123456 "[##-##-##]" -> "[12-34-56]"]
  - [format_int 1234 "##;(##)" -> "1234"]
  - [format_int (-1234) "##;(##)" -> "(1234)"]
*)

(** [format_int n fmt]は，書式指定文字列[fmt]の指示にしたがって，整数[n]を文字列に変換します。
*)
val format_int : int -> string -> string

(** [format_float n fmt]は，書式指定文字列[fmt]の指示にしたがって，実数[n]を文字列に変換します。
*)
val format_float : float -> string -> string


(** {6 日付/時刻の書式指定}
  以下の書式指定文字列を使用し，日付/時刻の書式を指定します。

  {10 【利用可能な書式指定文字】}
  - [YY ]            ２桁年（０パディングする）
  - [YYYY ]          ４桁年（０パディングする）
  - [M ]             月（０パディングしない）
  - [MM ]            月（０パディングする）
  - [B ]             月略名（Jan., Feb.,..., Dec.）
  - [BB ]            月正式名（１月, ２月,..., １２月）
  - [D ]             日（０パディングしない）
  - [DD ]            日（０パディングする）
  - [A ]             曜日略名（日, 月,..., 土）
  - [AA ]            曜日正式名（日曜日, 月曜日,..., 土曜日）
  - [G ]             年号略名（M, T, S, H）
  - [GG ]            年号正式名（明治, 大正, 昭和, 平成）
  - [E ]             和暦（０パディングしない）
  - [EE ]            和暦（０パディングする）
  - [h ]             時（２４時間表記，０パディングしない）
  - [hh ]            時（２４時間表記，０パディングする）
  - [H ]             時（１２時間表記，０パディングしない）
  - [HH ]            時（１２時間表記，０パディングする）
  - [m ]             分（０パディングしない）
  - [mm ]            分（０パディングする）
  - [s ]             秒（０パディングしない）
  - [ss ]            秒（０パディングする）
  - [t ]             午前/午後略名（AM, PM）
  - [tt ]            午前/午後正式名（午前, 午後）
  - [\<文字> ]       エスケープ文字
  - ['<文字列>', "<文字列>" ] リテラル文字列
  - [その他の文字 ]  結果の文字列にコピーされる。

  {10 【利用可能なエスケープ文字】}
  - [\b ]            バックスペース(BS)
  - [\n ]            行送り(LF)
  - [\r ]            復帰(CR)
  - [\t ]            水平タブ(TAB)
  - [\u<HHHH> ]      16進数値[HHHH]を持つUnicode文字
  - [\<文字> ]       文字自身

  {10 【使用例】}
  - [format_tm (tm_of_string "2010-10-23T15:21:10") "YYYY年MM月DD日" -> "2010年10月23日"]
  - [format_tm (Unix.localtime (Unix.time ())) "hh:mm:ss" -> "15:24:35"]
  - [format_tm (tm_of_string "1911-04-04") "GGEE年MM月DD日" -> "明治44年04月04日"]
  - [format_tm (tm_of_string "2040-01-01") "AA" -> "日曜日"]
*)

(** [format_tm tm fmt]は，書式指定文字列[fmt]の指示にしたがって，日付/時刻[tm]を文字列に変換します。
*)
val format_tm : Unix.tm -> string -> string

(** 文字列で表記した日付/時刻を解析して，[Unix.tm]に変換します。

  解析可能な文字列の書式は，以下のISO 8601(JIS X 0301)形式のサブセットです。
  - [YYYY['-'MM['-'DD]]'T'hh'[:'mm[':'ss]]]

  ただし，日付と時刻の区切り文字['T']の代わりに空白文字[' ']を使用することを許容します。

  {10 【使用例】}
  - [tm_of_string "2010-09-25" ->
      {tm_sec = 0; tm_min = 0; tm_hour = 0; tm_mday = 25; tm_mon = 8;
       tm_year = 110; tm_wday = 6; tm_yday = 267; tm_isdst = false}]
  - [tm_of_string "2010-09-23T10:12" ->
      {tm_sec = 0; tm_min = 12; tm_hour = 10; tm_mday = 23; tm_mon = 8;
       tm_year = 110; tm_wday = 4; tm_yday = 265; tm_isdst = false}]
 *)
val tm_of_string : string -> Unix.tm

(** [Unix.tm]をISO 8601(JIS X 0301)形式時刻文字列に変換します。*)
val string_of_tm : Unix.tm -> string

